// SPDX-License-Identifier: GPL-2.0
/*
 * Driver for the AK4490 DAC
 *
 * Author:	Aleksandr Chechulin <achechulin@gmail.com>
 *		Copyright 2021
 *
 * Based on code by __tkz__ <tkz@lrclk.com> (ak449x.c)
 * (https://github.com/tkztkztkz/linux/tree/npi-audio-4.11.y)
 */

#include <linux/delay.h>
#include <linux/gpio/consumer.h>
#include <linux/i2c.h>
#include <linux/module.h>
#include <linux/of_device.h>
#include <linux/of_gpio.h>
#include <linux/pm_runtime.h>
#include <linux/slab.h>
#include <linux/clk.h>
#include <sound/initval.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <sound/tlv.h>

#include "ak4490.h"

/*
#undef dev_dbg
#define dev_dbg dev_info
*/

/* Codec Private Data */
struct ak449x_priv {
	struct regmap *regmap;
	struct gpio_desc *reset_gpiod;
	struct gpio_desc *mute_gpiod;
	int digfil;			/* SSLOW, SD, SLOW bits */
	int fs;				/* Sampling rate */
	int fmt;
	int slots;
	int slot_width;
	int mute;
	struct timer_list timer;	/* UnMute timer */
};

static int    ak449x_reset = -1;

/* reg_default for ak4490 */
static const struct reg_default ak4490_reg_defaults[] = {
	{ 0x00, 0x04 },	/*	0x00	AK449X_00_CONTROL1	*/
	{ 0x01, 0x22 },	/*	0x01	AK449X_01_CONTROL2	*/
	{ 0x02, 0x00 },	/*	0x02	AK449X_02_CONTROL3	*/
	{ 0x03, 0xFF },	/*	0x03	AK449X_03_LCHATT	*/
	{ 0x04, 0xFF },	/*	0x04	AK449X_04_RCHATT	*/
	{ 0x05, 0x00 },	/*	0x05	AK449X_05_CONTROL4	*/
	{ 0x06, 0x00 },	/*	0x06	AK449X_06_DSD1		*/
	{ 0x07, 0x00 },	/*	0x07	AK449X_07_CONTROL5	*/
	{ 0x08, 0x00 },	/*	0x08	AK449X_08_SOUND_CONTROL	*/
	{ 0x09, 0x00 },	/*	0x09	AK449X_09_DSD2		*/
};

/*
 * Volume control:
 * from -127 to 0 dB in 0.5 dB steps (mute instead of -127.5 dB)
 */
static DECLARE_TLV_DB_SCALE(dac_tlv, -12750, 50, 1);

/*
 * DEM1 bit DEM0 bit Mode
 * 0 0 44.1kHz
 * 0 1 OFF (default)
 * 1 0 48kHz
 * 1 1 32kHz
 */
static const char * const ak449x_dem_select_texts[] = {
	"44.1kHz", "OFF", "48kHz", "32kHz"
};

/*
 * SSLOW, SD, SLOW bits Digital Filter Setting
 * 0, 0, 0 : Sharp Roll-Off Filter
 * 0, 0, 1 : Slow Roll-Off Filter
 * 0, 1, 0 : Short delay Sharp Roll-Off Filter
 * 0, 1, 1 : Short delay Slow Roll-Off Filter
 * 1, *, * : Super Slow Roll-Off Filter(NOS)
 */
static const char * const ak449x_digfil_select_texts[] = {
	"Sharp Roll-Off Filter",
	"Slow Roll-Off Filter",
	"Short delay Sharp Roll-Off Filter",
	"Short delay Slow Roll-Off Filter",
	"Super Slow Roll-Off Filter(NOS)"
};

/*
 * DZFB: Inverting Enable of DZF
 * 0: DZF goes H at Zero Detection
 * 1: DZF goes L at Zero Detection
 */
static const char * const ak449x_dzfb_select_texts[] = {"H", "L"};

/*
 * SC2-0 bits: Sound Mode Setting
 * ak4490 0-2
 */
static const char * const ak4490_sc_select_texts[] = {
	"Mode1", "Mode2", "Mode3"
};

/*
 * GC2-0 bits: Gain Control
 *           PCM    DSD    DSD BP
 * 0, 0, 0 : 2.8Vpp 2.6Vpp 2.5Vpp
 * 0, 0, 1 : 2.8Vpp 2.5Vpp 2.5Vpp
 * 0, 1, 0 : 2.5Vpp 2.5Vpp 2.5Vpp
 * 0, 1, 1 : 2.5Vpp 2.5Vpp 2.5Vpp Dup
 * 1, 0, 0 : 3.75Vpp 3.75Vpp 2.5Vpp
 * 1, 0, 1 : 3.75Vpp 2.5Vpp 2.5Vpp
 * 1, 1, 0 : 2.8Vpp 2.5Vpp 2.5Vpp Dup
 * 1, 1, 1 : 2.8Vpp 2.5Vpp 2.5Vpp Dup
 */

static const char * const ak449x_gc_select_texts[] = {
	"PCM/DSD2.8Vpp/DSDBP2.5Vpp",
	"PCM2.8Vpp/DSD2.5Vpp",
	"ALL2.5Vpp",
	"ALL2.5Vpp",
	"PCM/DSD3.75Vpp/DSDBP2.5Vpp",
	"PCM3.75Vpp/DSD2.5Vpp",
};

// generic control
static const struct soc_enum ak449x_dem_enum =
	SOC_ENUM_SINGLE(AK449X_01_CONTROL2, 1,
			ARRAY_SIZE(ak449x_dem_select_texts),
			ak449x_dem_select_texts);

static const struct soc_enum ak449x_digfil_enum =
	SOC_ENUM_SINGLE_EXT(ARRAY_SIZE(ak449x_digfil_select_texts),
			    ak449x_digfil_select_texts);

static const struct soc_enum ak449x_dzfb_enum =
	SOC_ENUM_SINGLE(AK449X_02_CONTROL3, 2,
			ARRAY_SIZE(ak449x_dzfb_select_texts),
			ak449x_dzfb_select_texts);

// chip specific control
static const struct soc_enum ak4490_sm_enum =
	SOC_ENUM_SINGLE(AK449X_08_SOUND_CONTROL, 0,
			ARRAY_SIZE(ak4490_sc_select_texts),
			ak4490_sc_select_texts);

static const struct soc_enum ak449x_gc_enum =
	SOC_ENUM_SINGLE(AK449X_07_CONTROL5, 1,
			ARRAY_SIZE(ak449x_gc_select_texts),
			ak449x_gc_select_texts);

static int get_digfil(struct snd_kcontrol *kcontrol,
		      struct snd_ctl_elem_value *ucontrol)
{
	struct snd_soc_component *component = snd_soc_kcontrol_component(kcontrol);
	struct ak449x_priv *ak449x = snd_soc_component_get_drvdata(component);

	ucontrol->value.enumerated.item[0] = ak449x->digfil;

	return 0;
}

static int set_digfil(struct snd_kcontrol *kcontrol,
		      struct snd_ctl_elem_value *ucontrol)
{
	struct snd_soc_component *component = snd_soc_kcontrol_component(kcontrol);
	struct ak449x_priv *ak449x = snd_soc_component_get_drvdata(component);
	int num;

	num = ucontrol->value.enumerated.item[0];
	if (num > 4)
		return -EINVAL;

	ak449x->digfil = num;

	/* write SD bit */
	snd_soc_component_update_bits(component, AK449X_01_CONTROL2,
			    AK449X_SD_MASK,
			    ((ak449x->digfil & 0x02) << 4));

	/* write SLOW bit */
	snd_soc_component_update_bits(component, AK449X_02_CONTROL3,
			    AK449X_SLOW_MASK,
			    (ak449x->digfil & 0x01));

	/* write SSLOW bit */
	snd_soc_component_update_bits(component, AK449X_05_CONTROL4,
			    AK449X_SSLOW_MASK,
			    ((ak449x->digfil & 0x04) >> 2));

	return 0;
}

static struct snd_kcontrol_new ak4490_snd_controls[] = {
	SOC_DOUBLE_R_TLV("Digital Playback Volume", AK449X_03_LCHATT, AK449X_04_RCHATT, 0, 0xFF, 0, dac_tlv),
	SOC_ENUM("De-emphasis", ak449x_dem_enum),
	SOC_ENUM_EXT("DigiFilt", ak449x_digfil_enum, get_digfil, set_digfil),
	SOC_ENUM("SoundMode", ak4490_sm_enum),
};

/* ak449x dapm widgets */
static const struct snd_soc_dapm_widget ak449x_dapm_widgets[] = {
	SND_SOC_DAPM_DAC("DACL", NULL, SND_SOC_NOPM, 0, 0),
	SND_SOC_DAPM_DAC("DACR", NULL, SND_SOC_NOPM, 0, 0),

	SND_SOC_DAPM_OUTPUT("OUTL"),
	SND_SOC_DAPM_OUTPUT("OUTR"),
};

static const struct snd_soc_dapm_route ak449x_intercon[] = {
	{ "DACL", NULL, "Playback" },
	{ "DACR", NULL, "Playback" },
	{ "OUTL", NULL, "DACL" },
	{ "OUTR", NULL, "DACR" },
};

static int ak449x_rstn_control(struct snd_soc_component *component, int bit)
{
	int ret;

	if (bit)
		ret = snd_soc_component_update_bits(component,
					  AK449X_00_CONTROL1,
					  AK449X_RSTN_MASK,
					  0x1);
	else
		ret = snd_soc_component_update_bits(component,
					  AK449X_00_CONTROL1,
					  AK449X_RSTN_MASK,
					  0x0);
	return ret;
}

static int ak449x_hw_params(struct snd_pcm_substream *substream,
			    struct snd_pcm_hw_params *params,
			    struct snd_soc_dai *dai)
{
	struct snd_soc_component *component = dai->component;
	struct ak449x_priv *ak449x = snd_soc_component_get_drvdata(component);
	int pcm_width = max(params_physical_width(params), ak449x->slot_width);
	int nfs1;
	u8 format;

	nfs1 = params_rate(params);
	ak449x->fs = nfs1;

	dev_dbg(component->dev, "%s pcm_width = %d, fmt = %x\n", __FUNCTION__, pcm_width, ak449x->fmt);

	/* Master Clock Frequency Auto Setting Mode Enable */
	snd_soc_component_update_bits(component, AK449X_00_CONTROL1, 0x80, 0x80);

	switch (pcm_width) {
	case 16:
		if (ak449x->fmt == SND_SOC_DAIFMT_I2S)
			format = AK449X_DIF_24BIT_I2S;
		else
			format = AK449X_DIF_16BIT_LSB;
		break;
	case 24:
		switch (ak449x->fmt) {
		case SND_SOC_DAIFMT_I2S:
			format = AK449X_DIF_24BIT_I2S;
			break;
		case SND_SOC_DAIFMT_RIGHT_J:
			format = AK449X_DIF_24BIT_LSB;
			break;
		default:
			return -EINVAL;
		}
		break;
	case 32:
		switch (ak449x->fmt) {
		case SND_SOC_DAIFMT_I2S:
			format = AK449X_DIF_32BIT_I2S;
			break;
		case SND_SOC_DAIFMT_LEFT_J:
			format = AK449X_DIF_32BIT_MSB;
			break;
		case SND_SOC_DAIFMT_RIGHT_J:
			format = AK449X_DIF_32BIT_LSB;
			break;
		case SND_SOC_DAIFMT_DSP_B:
			format = AK449X_DIF_32BIT_MSB;
			break;
		default:
			return -EINVAL;
		}
		break;
	default:
		return -EINVAL;
	}

	snd_soc_component_update_bits(component, AK449X_00_CONTROL1, AK449X_DIF_MASK, format);
	dev_dbg(component->dev, "%s set dif format = %x\n", __FUNCTION__, format);

	ak449x_rstn_control(component, 0);
	ak449x_rstn_control(component, 1);

	return 0;
}

static int ak449x_set_dai_fmt(struct snd_soc_dai *dai, unsigned int fmt)
{
	struct snd_soc_component *component = dai->component;
	struct ak449x_priv *ak449x = snd_soc_component_get_drvdata(component);

	dev_dbg(component->dev, "%s fmt = %d\n", __FUNCTION__, fmt);

	switch (fmt & SND_SOC_DAIFMT_MASTER_MASK) {
	case SND_SOC_DAIFMT_CBS_CFS: /* Slave Mode */
	case SND_SOC_DAIFMT_CBM_CFS: /* Clock Slave and Frame Master */
		break;
	case SND_SOC_DAIFMT_CBM_CFM: /* Master Mode is not supported */
	case SND_SOC_DAIFMT_CBS_CFM:
	default:
		dev_warn(component->dev, "Master mode required external I2S master\n");
	}

	switch (fmt & SND_SOC_DAIFMT_FORMAT_MASK) {
	case SND_SOC_DAIFMT_I2S:
	case SND_SOC_DAIFMT_LEFT_J:
	case SND_SOC_DAIFMT_RIGHT_J:
	case SND_SOC_DAIFMT_DSP_B:
		ak449x->fmt = fmt & SND_SOC_DAIFMT_FORMAT_MASK;
		break;
	default:
		dev_err(component->dev, "Audio format 0x%02X unsupported\n",
			fmt & SND_SOC_DAIFMT_FORMAT_MASK);
		return -EINVAL;
	}

	return 0;
}

static void timer_callback(struct timer_list *t)
{
	struct ak449x_priv *ak449x = from_timer(ak449x, t, timer);

	if (!ak449x->mute) {
		gpiod_set_value_cansleep(ak449x->mute_gpiod, 1);
	}
}

static int ak449x_set_dai_mute(struct snd_soc_dai *dai, int mute, int stream)
{
	struct snd_soc_component *component = dai->component;
	struct ak449x_priv *ak449x = snd_soc_component_get_drvdata(component);
	int nfs, ndt, ret;

	/* state check */
	if ( ak449x->mute == mute ) {
		return 0;
	}
	ak449x->mute = mute;

	dev_dbg(component->dev, "%s mute = %d\n", __FUNCTION__, mute);

	/* calculate att transition time */
	nfs = ak449x->fs;
	ndt = 7424 / (nfs / 1000);

	if (mute) {
		ret = snd_soc_component_update_bits(component, AK449X_01_CONTROL2, 0x01, 1);
		if (ak449x->mute_gpiod)
			gpiod_set_value_cansleep(ak449x->mute_gpiod, 0);
	} else {
		ret = snd_soc_component_update_bits(component, AK449X_01_CONTROL2, 0x01, 0);
		if (ak449x->mute_gpiod)
			mod_timer(&ak449x->timer, jiffies + msecs_to_jiffies(ndt));
	}

	return 0;
}

static void ak449x_power_off(struct device *dev)
{
	struct ak449x_priv *ak449x = dev_get_drvdata(dev);

	if (ak449x_reset != 0 && ak449x->reset_gpiod) {
		ak449x_reset = 0;
		dev_dbg(dev, "%s\n", __FUNCTION__);
		gpiod_set_value_cansleep(ak449x->reset_gpiod, 0); // pdn enable
		usleep_range(1000, 2000);
	}
}

static void ak449x_power_on(struct device *dev)
{
	struct ak449x_priv *ak449x = dev_get_drvdata(dev);

	if (ak449x_reset != 1 && ak449x->reset_gpiod) {
		ak449x_reset = 1;
		dev_dbg(dev, "%s\n", __FUNCTION__);
		gpiod_set_value_cansleep(ak449x->reset_gpiod, 1); // pdn disable
		usleep_range(1000, 2000);
	}
}

static int ak449x_startup(struct snd_pcm_substream *substream, struct snd_soc_dai *dai)
{
	struct snd_soc_component *component = dai->component;
	struct ak449x_priv *ak449x = snd_soc_component_get_drvdata(component);
	int err;

	dev_dbg(component->dev, "%s\n", __FUNCTION__);

	ak449x->mute = -1; /* reset mute state */

	if (ak449x_reset == -1 )
		ak449x_power_off(component->dev);
	ak449x_power_on(component->dev);

	snd_soc_component_read(component, AK449X_00_CONTROL1); /* dummy read */
	err = snd_soc_component_update_bits(component, AK449X_00_CONTROL1, 0x80, 0x80);   /* ACKS bit = 1; 10000000 */
	err = snd_soc_component_update_bits(component, AK449X_00_CONTROL1, 0x80, 0x80);   /* ACKS bit = 1; 10000000 */

	if (err < 0) {
		dev_err(component->dev, "i2c failed");
		return err;
	}

	ak449x_rstn_control(component, 1);

	return 0;
}

static struct snd_soc_dai_ops ak449x_dai_ops = {
	.startup	= ak449x_startup,
	.hw_params	= ak449x_hw_params,
	.set_fmt	= ak449x_set_dai_fmt,
	.mute_stream	= ak449x_set_dai_mute,
};

static struct snd_soc_dai_driver ak449x_dai = {
	.name = "ak449x-aif",
	.playback = {
		.stream_name =  "Playback",
		.channels_min = 2,
		.channels_max = 2,
		.rates =        SNDRV_PCM_RATE_CONTINUOUS,
		.rate_min =     8000,
		.rate_max =	1536000,
		.formats =	SNDRV_PCM_FMTBIT_S16_LE |
			 	SNDRV_PCM_FMTBIT_S24_LE |
			 	SNDRV_PCM_FMTBIT_S32_LE,
	},
	.ops = &ak449x_dai_ops,
};

#ifdef CONFIG_PM
static int __maybe_unused ak449x_runtime_suspend(struct device *dev)
{
	struct ak449x_priv *ak449x = dev_get_drvdata(dev);

        dev_dbg(dev, "%s\n", __FUNCTION__);

	regcache_cache_only(ak449x->regmap, true);

	if (ak449x_reset) {
		ak449x_power_off(dev);
	}

	return 0;
}

static int __maybe_unused ak449x_runtime_resume(struct device *dev)
{
	struct ak449x_priv *ak449x = dev_get_drvdata(dev);

        dev_dbg(dev, "%s\n", __FUNCTION__);

	if (ak449x_reset != 1) {
		ak449x_power_on(dev);
	}

	regcache_cache_only(ak449x->regmap, false);
	regcache_mark_dirty(ak449x->regmap);

	return regcache_sync(ak449x->regmap);
}
#endif /* CONFIG_PM */

static const struct snd_soc_component_driver ak449x_component_driver = {
	.controls		= ak4490_snd_controls,
	.num_controls		= ARRAY_SIZE(ak4490_snd_controls),
	.dapm_widgets		= ak449x_dapm_widgets,
	.num_dapm_widgets	= ARRAY_SIZE(ak449x_dapm_widgets),
	.dapm_routes		= ak449x_intercon,
	.num_dapm_routes	= ARRAY_SIZE(ak449x_intercon),
};

static struct regmap_config ak4490_regmap = {
	.reg_bits = 8,
	.val_bits = 8,
	.max_register = AK449X_09_DSD2,
	.reg_defaults = ak4490_reg_defaults,
	.num_reg_defaults = ARRAY_SIZE(ak4490_reg_defaults),
	.cache_type = REGCACHE_RBTREE,
};

static const struct dev_pm_ops ak449x_pm = {
	SET_RUNTIME_PM_OPS(ak449x_runtime_suspend, ak449x_runtime_resume, NULL)
};

static int ak449x_probe(struct device *dev, struct regmap *regmap)
{
	struct ak449x_priv *ak449x;
	int ret;

	dev_dbg(dev, "%s\n", __FUNCTION__);
	ak449x = devm_kzalloc(dev, sizeof(*ak449x), GFP_KERNEL);
	if (!ak449x)
		return -ENOMEM;

	dev_set_drvdata(dev, ak449x);
	ak449x->regmap = regmap;

	// Setup gpios
	ak449x->reset_gpiod = devm_gpiod_get_optional(dev, "reset", GPIOD_OUT_LOW);
	if (IS_ERR(ak449x->reset_gpiod)){
		dev_err(dev, "error requesting reset-gpios: %ld\n", PTR_ERR(ak449x->reset_gpiod));
		return PTR_ERR(ak449x->reset_gpiod);
	}

	ak449x->mute_gpiod = devm_gpiod_get_optional(dev, "mute", GPIOD_OUT_LOW);
	if (IS_ERR(ak449x->mute_gpiod)) {
		dev_err(dev, "error requesting mute-gpios: %ld\n", PTR_ERR(ak449x->mute_gpiod));
		return PTR_ERR(ak449x->mute_gpiod);
	}

	timer_setup(&ak449x->timer, timer_callback, 0);

	// Register codec
	ret =  devm_snd_soc_register_component(dev, &ak449x_component_driver, &ak449x_dai, 1);
	if (ret != 0) {
		dev_err(dev, "Failed to register CODEC: %d\n", ret);
		return ret;
	}

	pm_runtime_set_active(dev);
	pm_runtime_enable(dev);
	pm_runtime_idle(dev);

	dev_info(dev, "Registered ak4490");
	return 0;
}

static int ak449x_remove(struct device *dev)
{
	pm_runtime_disable(dev);

	return 0;
}

static int ak449x_i2c_probe(struct i2c_client *i2c)
{
	struct regmap *regmap;

	regmap = devm_regmap_init_i2c(i2c, &ak4490_regmap);
	if (IS_ERR(regmap))
		return PTR_ERR(regmap);

	return ak449x_probe(&i2c->dev, regmap);
}

static int ak449x_i2c_remove(struct i2c_client *i2c)
{
	ak449x_remove(&i2c->dev);

	return 0;
}

static const struct of_device_id ak449x_of_match[] = {
	{ .compatible = "asahi-kasei,ak4490", },
	{ },
};
MODULE_DEVICE_TABLE(of, ak449x_of_match);

static struct i2c_driver ak449x_i2c_driver = {
	.driver = {
		.name = "ak4490",
		.pm = &ak449x_pm,
		.of_match_table = ak449x_of_match,
		},
	.probe_new = ak449x_i2c_probe,
	.remove = ak449x_i2c_remove,
};

module_i2c_driver(ak449x_i2c_driver);

MODULE_DESCRIPTION("ASoC AK4490 codec driver - I2C");
MODULE_LICENSE("GPL v2");
