// SPDX-License-Identifier: GPL-2.0
/*
 * Card driver for the AK4490 DAC
 *
 * Author:	Aleksandr Chechulin <achechulin@gmail.com>
 *		Copyright 2021
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/gpio/consumer.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>

/* Card private data */
struct card_priv {
	struct gpio_desc *clock44;
	struct gpio_desc *clock48;
	unsigned long clock_div;	/* Clock divided at board, 1 or 2 */
};

enum dac_clock_type {
	DAC_CLOCK_CLK44EN,	/* Clock rate of CLK44EN attached to GPIO6 pin */
	DAC_CLOCK_CLK48EN,	/* Clock rate of CLK48EN attached to GPIO5 pin */
};

static const unsigned long dac_clock_rate[] = {
	22579200UL,
	24576000UL,
};

static void snd_dac_clk_select(struct snd_soc_card *card, enum dac_clock_type clk_id)
{
	struct card_priv *priv = snd_soc_card_get_drvdata(card);

	switch (clk_id) {
	case DAC_CLOCK_CLK44EN:
		gpiod_set_value_cansleep(priv->clock48, 0);
		gpiod_set_value_cansleep(priv->clock44, 1);
		break;
	case DAC_CLOCK_CLK48EN:
		gpiod_set_value_cansleep(priv->clock44, 0);
		gpiod_set_value_cansleep(priv->clock48, 1);
		break;
	}
}

static enum dac_clock_type snd_dac_clk_for_rate(int sample_rate)
{
	enum dac_clock_type ctype;

	switch (sample_rate) {
	case 11025:
	case 22050:
	case 44100:
	case 88200:
	case 176400:
	case 352800:
		ctype = DAC_CLOCK_CLK44EN;
		break;
	default:
		ctype = DAC_CLOCK_CLK48EN;
		break;
	}
	return ctype;
}

static int snd_dac_hw_params(struct snd_pcm_substream *substream, struct snd_pcm_hw_params *params)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_card *card = rtd->card;
	struct card_priv *priv = snd_soc_card_get_drvdata(card);

	enum dac_clock_type ctype = snd_dac_clk_for_rate(params_rate(params));
	int bclk_ratio;
	int ret;

	snd_dac_clk_select(card, ctype);

	bclk_ratio = (dac_clock_rate[ctype] / priv->clock_div) / params_rate(params);

	dev_dbg(card->dev, "%s ctype = %d, bclk_ratio = %d\n", __FUNCTION__, ctype, bclk_ratio);

	ret = snd_soc_dai_set_bclk_ratio(asoc_rtd_to_cpu(rtd, 0), bclk_ratio);
	if (ret < 0)
		dev_warn(card->dev, "snd_soc_dai_set_bclk_ratio(cpu) failed: %d\n", ret);

	return 0;
}

/* machine stream operations */
static struct snd_soc_ops snd_dac_ops = {
	.hw_params = snd_dac_hw_params,
};

SND_SOC_DAILINK_DEFS(snd_dai_defs,
	DAILINK_COMP_ARRAY(COMP_CPU("bcm2708-i2s.0")),
	DAILINK_COMP_ARRAY(COMP_CODEC("ak449x-aif", "ak449x-aif")),
	DAILINK_COMP_ARRAY(COMP_PLATFORM("bcm2708-i2s.0")));

static struct snd_soc_dai_link snd_dac_dai[] = {
{
	.name		= "AK449x",
	.stream_name	= "AK449x HiFi",
	SND_SOC_DAILINK_REG(snd_dai_defs),
	.dai_fmt	= SND_SOC_DAIFMT_I2S | SND_SOC_DAIFMT_NB_NF |
		SND_SOC_DAIFMT_CBM_CFS	/* codec clk provider & frame consumer */,
	.ops		= &snd_dac_ops,
},
};

/* audio machine driver */
static struct snd_soc_card snd_dac = {
	.name         = "AK449x",
	.driver_name  = "simple-ak449x",
	.owner        = THIS_MODULE,
	.dai_link     = snd_dac_dai,
	.num_links    = ARRAY_SIZE(snd_dac_dai),
};

static int snd_dac_probe(struct platform_device *pdev)
{
	int ret = 0;
	struct snd_soc_card *card = &snd_dac;
	struct card_priv *priv;
	u32 clock_div;

	priv = devm_kzalloc(&pdev->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	card->dev = &pdev->dev;
	snd_soc_card_set_drvdata(card, priv);

	// Setup DAI link
	if (pdev->dev.of_node) {
		struct snd_soc_dai_link *dai;
		struct device_node *i2s_node;
		struct device_node *codec_node;

		dai = &snd_dac_dai[0];

		i2s_node = of_parse_phandle(pdev->dev.of_node, "i2s-controller", 0);
		codec_node = of_parse_phandle(pdev->dev.of_node, "audio-codec", 0);

		if (i2s_node) {
			dai->cpus[0].dai_name = NULL;
			dai->cpus[0].of_node = i2s_node;
			dai->platforms[0].name = NULL;
			dai->platforms[0].of_node = i2s_node;
		} else {
			dev_err(&pdev->dev, "i2s-controller not specified\n");
		}

		if (codec_node) {
			dai->codecs[0].name = NULL;
			dai->codecs[0].of_node = codec_node;
		} else {
			dev_err(&pdev->dev, "audio-codec not specified\n");
		}
	}

	// Setup gpios
	priv->clock44 = devm_gpiod_get_optional(&pdev->dev, "clock44", GPIOD_OUT_LOW);
	if (IS_ERR(priv->clock44)){
		dev_err(&pdev->dev, "error requesting clock44-gpios: %ld\n", PTR_ERR(priv->clock44));
		return PTR_ERR(priv->clock44);
	}

	priv->clock48 = devm_gpiod_get_optional(&pdev->dev, "clock48", GPIOD_OUT_LOW);
	if (IS_ERR(priv->clock48)){
		dev_err(&pdev->dev, "error requesting clock48-gpios: %ld\n", PTR_ERR(priv->clock48));
		return PTR_ERR(priv->clock48);
	}

	priv->clock_div = 1;
	ret = of_property_read_u32(pdev->dev.of_node, "clock-div", &clock_div);
	if (!ret) {
		priv->clock_div = clock_div;
	}

	ret = devm_snd_soc_register_card(&pdev->dev, card);
	if (ret && ret != -EPROBE_DEFER)
		dev_err(&pdev->dev, "devm_snd_soc_register_card failed: %d\n", ret);

	return ret;
}

static int snd_dac_remove(struct platform_device *pdev)
{
	return snd_soc_unregister_card(&snd_dac);
}

static const struct of_device_id snd_dac_of_match[] = {
	{ .compatible = "simple-ak449x", },
	{},
};
MODULE_DEVICE_TABLE(of, snd_dac_of_match);

static struct platform_driver snd_dac_driver = {
	.driver = {
		.name   = "simple-ak449x",
		.owner  = THIS_MODULE,
		.of_match_table = snd_dac_of_match,
	},
	.probe	= snd_dac_probe,
	.remove = snd_dac_remove,
};

module_platform_driver(snd_dac_driver);

MODULE_DESCRIPTION("Simple AK449x card driver");
MODULE_LICENSE("GPL v2");

